#include <windows.h>
#include <stdio.h>
#include "FileInfo.h"


static unsigned short MPEG_SAMPLE_RATES[4][4] = {
  { 0, 0, 0, 0 }, // unknown
  { 44100, 48000, 32000, 0 }, // V1
  { 22050, 24000, 16000, 0 }, // V2
  { 11025, 12000, 8000, 0 } }; // V3
	
static unsigned short MPEG_BIT_RATES[4][4][16] = {
  {{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
   {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
   {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
   {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}},
	
  {{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
   {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,0},
   {0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,0},
   {0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,0}},
	
  {{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
   {0,32,48, 56, 64, 80, 96,112,128,144,160,176,192,224,256,0},
   {0, 8,16,24, 32, 40, 48, 56, 64, 80, 96, 112,128,144,160,0},
   {0, 8,16,24, 32, 40, 48, 56, 64, 80, 96, 112,128,144,160,0}},
	
  {{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
   {0,32,48, 56, 64, 80, 96,112,128,144,160,176,192,224,256,0},
   {0, 8,16,24, 32, 40, 48, 56, 64, 80, 96, 112,128,144,160,0},
   {0, 8,16,24, 32, 40, 48, 56, 64, 80, 96, 112,128,144,160,0}}
};
	
typedef enum XINGHDRCONST {
  XH_FRAMES_FLAG = 1,
  XH_BYTES_FLAG = 2,
  XH_TOC_FLAG = 4,
  XH_VBR_SCALE_FLAG = 8
} XINGHDRCONST;

static int FrameHeaderValid(MPEGDATA *m)
{
  return ((m->fileLen>5) && (m->ver>0) && (m->layer>0) && (m->bitRate!=0)
	  && (m->sampleRate>0))?TRUE:FALSE;
}

static unsigned long CalcFrameLength(int layer, int sampleRate, int bitRate,
				     int padding)
{
  if (sampleRate > 0) {
    if (layer == 1)
      return (12 * bitRate * 1000 / sampleRate + (padding*4));
    else
      return (144 * bitRate * 1000 / sampleRate + padding);
  } else
    return 0;
}

static int DecodeHeader(unsigned long MPEGHeader, MPEGDATA *m)
{
  unsigned int s;

  s = m->fileLen;
  memset(m, 0, sizeof(MPEGDATA));
  m->fileLen = s;

  if ((MPEGHeader & 0xffe00000) == 0xffe00000) {
    switch((MPEGHeader >> 19) & 0x3) {
    case 0 : m->ver = MPEG_VERSION_25; break;
    case 1 : m->ver = MPEG_VERSION_UNKNOWN; break;
    case 2 : m->ver = MPEG_VERSION_2; break;
    case 3 : m->ver = MPEG_VERSION_1; break;
    }
		
    if (m->ver != MPEG_VERSION_UNKNOWN) {
      m->layer = 4 - ((MPEGHeader >> 17) & 0x3);
      if (m->layer > 3) m->layer = 0;
			
      m->sampleRate = MPEG_SAMPLE_RATES[m->ver][((MPEGHeader >> 10) & 0x3)];
      m->errorProt = ((MPEGHeader >> 16) & 1) == 1;
      m->copyright = ((MPEGHeader >> 3) & 1) == 1;
      m->original = ((MPEGHeader >> 2) & 1) == 1;
      m->mode = ((MPEGHeader >> 6) & 3);
      m->padding = ((MPEGHeader >> 9) & 1) == 1;
      m->bitRate = MPEG_BIT_RATES[m->ver][m->layer][((MPEGHeader >> 12)&0xf)];
			
      if (m->bitRate == 0)  m->duration = 0;
      else m->duration = (m->fileLen*8) / (m->bitRate*1000);
			
      m->frameLen = CalcFrameLength(m->layer, m->sampleRate, m->bitRate,
				    m->padding);
    }
    return FrameHeaderValid(m);
  } else
    return FALSE;
}


static unsigned long GetLong(FILE *fp)
{
  unsigned long a;
  a = fgetc(fp);
  a = a<<8;
  a |= fgetc(fp)&255;
  a = a<<8;
  a |= fgetc(fp)&255;
  a = a<<8;
  a |= fgetc(fp)&255;
  return a;
}

int GetMP3Info(FILE *fp, MPEGDATA *data)
{
  unsigned long mp3hdr;
  long dataLen;
  unsigned long flags, frames, dev, firstFramePos;
	
  memset(data, 0, sizeof(MPEGDATA));
	
  if (!fp) return FALSE;
  fseek(fp, 0, SEEK_END);
  data->fileLen = ftell(fp);
  fseek(fp, 0, SEEK_SET);
  if (data->fileLen<=5) return FALSE;
	
  firstFramePos = 0;

  mp3hdr = GetLong(fp);
  while (!DecodeHeader(mp3hdr, data) && !feof(fp) && ftell(fp)<6000) {
    fseek(fp, SEEK_CUR, -3);
    mp3hdr = GetLong(fp);
  }
	
  firstFramePos= ftell(fp)-4;
  dataLen = (data->fileLen - firstFramePos- data->frameLen +
	     (data->errorProt?2:0));
  if ( dataLen<0 || feof(fp) || !DecodeHeader(mp3hdr, data) ) return FALSE;
	
  fseek(fp, firstFramePos+data->frameLen, SEEK_SET);
  mp3hdr = GetLong(fp);
	
  if (!DecodeHeader(mp3hdr, data)) return FALSE;
	
  if (data->ver== 1) {
    if (data->mode !=3) dev = 36;
    else dev = 21;
  } else {
    if (data->mode != 3) dev = 21;
    else dev = 13;
  }
	
  fseek(fp, firstFramePos+dev, SEEK_SET);
  mp3hdr = GetLong(fp);
  if (mp3hdr=='XING') {
    data->bitRate = 0xffffffff;
    flags = GetLong(fp);
    if (flags & XH_FRAMES_FLAG)	frames = GetLong(fp);
    else frames = 1;
    data->duration = (unsigned long)((1152.0 / (double)data->sampleRate)
				     * (double)frames);
    data->frameLen = (unsigned long)((double)data->fileLen / (double)frames);
  }
	
  if (data->bitRate==0xffffffff)
    strcpy(data->bitRateText, "VBR");
  else
    sprintf(data->bitRateText, "%dKbps", data->bitRate);
	
  if (data->duration<60)
    sprintf(data->durationText, "0:%02d", data->duration);
  else if (data->duration<3600)
    sprintf(data->durationText, "%d:%02d", data->duration/60,
	    data->duration%60);
  else
    sprintf(data->durationText, "%d:%02d:%02d", data->duration/3600,
	    (data->duration%3600)/60, (data->duration%3600)%60);
	
  return TRUE;
}
