#include <windows.h>
#include <stdio.h>
#include "resource.h"
#include "plugin.h"
#include "FileInfo.h"


static HINSTANCE hInstance = NULL;
static POINT position = {0, 0};
static DWORD enabled = FALSE;


BOOL APIENTRY DllMain( HANDLE hModule, DWORD reason, LPVOID resvd )
{
  hInstance = (HINSTANCE)hModule;
  return TRUE;
}


static void FormatString(char *in, char *fname, char *out, int len,
			 char *title, char *artist, char *album,
			 char *year, int track, char *bitrate, char *duration)
{
  int i, j, slen;
  char obuff[8192];
  char buff[8];
	
  if (in[0]==0) { strncpy(out, fname, len); return; }

  slen = strlen(in);
  for (i=j=0; i<slen && j<8000; i++) {
    obuff[j] = 0;
    if (!strncmp(in+i, "%title%", 7)) {
      strcat(obuff+j, title);
      i += 6;
      j += strlen(title);
    } else if (!strncmp(in+i, "%artist%", 8)) {
      strcat(obuff+j, artist);
      i += 7;
      j += strlen(artist);
    } else if (!strncmp(in+i, "%album%", 7)) {
      strcat(obuff+j, album);
      i += 6;
      j += strlen(album);
    } else if (!strncmp(in+i, "%year%", 6)) {
      strcat(obuff+j, year);
      i += 5;
      j += strlen(year);
    } else if (!strncmp(in+i, "%bitrate%", 9)) {
      strcat(obuff+j, bitrate);
      i += 8;
      j += strlen(bitrate);
    } else if (!strncmp(in+i, "%duration%", 10)) {
      strcat(obuff+j, duration);
      i += 9;
      j += strlen(duration);
    } else if (!strncmp(in+i, "%track%", 7)) {
      if (track>0)
	sprintf(buff, "%2d", track);
      else
	buff[0] = 0;
      strcat(obuff+j, buff);
      i += 6;
      j += strlen(buff);
    } else
      obuff[j++] = in[i];
  }
  obuff[j]=0;
  strncpy(out, obuff, len);
  out[len-1] = 0;
}

static void CenterDialog( HWND hwndDlg )
{
  HWND hwndOwner; 
  RECT rc, rcDlg, rcOwner; 

  hwndOwner = GetParent(hwndDlg);
  if (hwndOwner == NULL) hwndOwner = GetDesktopWindow(); 
  GetWindowRect(hwndOwner, &rcOwner); 
  GetWindowRect(hwndDlg, &rcDlg); 
  CopyRect(&rc, &rcOwner); 
  OffsetRect(&rcDlg, -rcDlg.left, -rcDlg.top); 
  OffsetRect(&rc, -rc.left, -rc.top); 
  OffsetRect(&rc, -rcDlg.right, -rcDlg.bottom); 
  SetWindowPos(hwndDlg, HWND_TOP, rcOwner.left + (rc.right / 2),
	       rcOwner.top + (rc.bottom / 2), 0, 0, SWP_NOSIZE); 
}

static void PositionWindow( HWND hdlg, POINT *pt)
{
  HDC hdc;
  if (pt->x!=0 && pt->y!=0) {
    hdc=CreateIC("DISPLAY", "", "", NULL);
    pt->x = min( GetDeviceCaps(hdc, HORZRES)-100, pt->x);
    pt->x = max( pt->x, 0 );
    pt->y = min( GetDeviceCaps(hdc, VERTRES)-100, pt->y);
    pt->y = max( pt->y, 0 );
    DeleteDC(hdc);
    SetWindowPos(hdlg, HWND_TOP, pt->x, pt->y, 0,0, SWP_NOSIZE);
  } else
    CenterDialog(hdlg);
}

static void SaveWindowPosition( HWND hdlg, POINT *pt )
{
  RECT saverect;
  GetWindowRect(hdlg, &saverect);
  pt->x = saverect.left;
  pt->y = saverect.top;
}

EXPORT int FileFilterInitialize( MINIFILEPARAM *param )
{
  DWORD type, len;

  // First, check that we can grok this plugin version!
  if (!param) return FALSE;
  if (param->version != PLUGINVERSION) return FALSE;

  strncpy(param->title, "MP3 ID3 File Renamer", MAXTITLE);
  param->version = PLUGINVERSION;
  param->user = calloc(512, 1);
  strcpy((char*)param->user, "%title% - %artist%");
  strcpy((char*)param->user+256, "%duration%");

  len = 256;
  type = REG_SZ;
  RegQueryValueEx(param->userKey, "string", 0, &type,
		  (unsigned char *)param->user, &len);

  len = 256;
  type = REG_SZ;
  RegQueryValueEx(param->userKey, "commstring", 0, &type,
		  (unsigned char *)param->user+256, &len);

  type = REG_BINARY;
  len = sizeof(POINT);
  RegQueryValueEx(param->userKey, "position", 0, &type,
		  (unsigned char *)&position, &len);

  type = REG_DWORD;
  len = sizeof(DWORD);
  RegQueryValueEx(param->userKey, "enabled", 0, &type,
		  (unsigned char *)&enabled, &len);
  return TRUE;
}


static BOOL CALLBACK Dialog(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  static MINIFILEPARAM *param;
  static int last = 0;
  char buff[256], outBuff[256];
  int en;

  switch(msg) {
  case WM_USER:
    en = (SendDlgItemMessage(hDlg, IDC_ENABLE, BM_GETCHECK, 0, 0)==BST_CHECKED)
      ?TRUE:FALSE;
    EnableWindow(GetDlgItem(hDlg, IDC_EDIT), en);
    EnableWindow(GetDlgItem(hDlg, IDC_EDITCOMM), en);
    EnableWindow(GetDlgItem(hDlg, IDC_TITLE), en);
    EnableWindow(GetDlgItem(hDlg, IDC_ALBUM), en);
    EnableWindow(GetDlgItem(hDlg, IDC_ARTIST), en);
    EnableWindow(GetDlgItem(hDlg, IDC_BITRATE), en);
    EnableWindow(GetDlgItem(hDlg, IDC_DURATION), en);
    EnableWindow(GetDlgItem(hDlg, IDC_YEAR), en);
    EnableWindow(GetDlgItem(hDlg, IDC_TRACK), en);
    EnableWindow(GetDlgItem(hDlg, IDC_STATIC1), en);
    EnableWindow(GetDlgItem(hDlg, IDC_STATIC2), en);
    EnableWindow(GetDlgItem(hDlg, IDC_SAMPLE), en);
    EnableWindow(GetDlgItem(hDlg, IDC_SAMPLECOMM), en);
    EnableWindow(GetDlgItem(hDlg, IDC_STATICN), en);
    EnableWindow(GetDlgItem(hDlg, IDC_STATICC), en);
    if (en) {
      GetDlgItemText(hDlg, IDC_EDIT, buff, 64);
      FormatString(buff, "Filename.mp3", outBuff, 256, "Song Title",
		   "The Band", "Red Album", "1969", 3, "128Kbps", "3:36");
      SetDlgItemText(hDlg, IDC_SAMPLE, outBuff);

      GetDlgItemText(hDlg, IDC_EDITCOMM, buff, 64);
      FormatString(buff, "1.43MB", outBuff, 256, "Song Title", "The Band",
		   "Red Album", "1969", 3, "128Kbps", "3:36");
      SetDlgItemText(hDlg, IDC_SAMPLECOMM, outBuff);
    } else {
      SetDlgItemText(hDlg, IDC_SAMPLE, "Filename.mp3");
      SetDlgItemText(hDlg, IDC_SAMPLECOMM, "1.43MB");
    }
    return TRUE;

  case WM_INITDIALOG:
    param = (MINIFILEPARAM*)lParam;
    PositionWindow(hDlg, &position);
    SendDlgItemMessage(hDlg, IDC_ENABLE, BM_SETCHECK, 
		       enabled?BST_CHECKED:BST_UNCHECKED, 0);
    SetDlgItemText(hDlg, IDC_EDIT, (char*)param->user);
    SetDlgItemText(hDlg, IDC_EDITCOMM, (char*)param->user+256);
    SetFocus(GetDlgItem(hDlg, enabled?IDC_EDIT:IDC_ENABLE));
    SendMessage(hDlg, WM_USER, 0, 0);
    last = IDC_EDIT;
    return FALSE;


  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDOK:
      SaveWindowPosition(hDlg, &position);
      GetDlgItemText(hDlg, IDC_EDIT, (char*)param->user, 256);
      GetDlgItemText(hDlg, IDC_EDITCOMM, ((char*)param->user)+256, 256);
      if (SendDlgItemMessage(hDlg, IDC_ENABLE, BM_GETCHECK, 0, 0)==BST_CHECKED)
	enabled = TRUE;
      else
	enabled = FALSE;
      EndDialog(hDlg, TRUE);
      return TRUE;
		
    case IDCANCEL:
      SaveWindowPosition(hDlg, &position);
      EndDialog(hDlg, FALSE);
      return TRUE;
		
    case IDC_EDIT:
      last = IDC_EDIT;
      return SendMessage(hDlg, WM_USER, 0, 0);

    case IDC_EDITCOMM:
      last = IDC_EDITCOMM;
      return SendMessage(hDlg, WM_USER, 0, 0);

    case IDC_ENABLE:
      SendMessage(hDlg, WM_USER, 0, 0);
      return TRUE;

    case IDC_TITLE:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%title%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_ARTIST:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%artist%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_ALBUM:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%album%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_YEAR:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%year%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_TRACK:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%track%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_BITRATE:
      SendDlgItemMessage(hDlg, last, EM_REPLACESEL, FALSE, (long)"%bitrate%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_DURATION:
      SendDlgItemMessage(hDlg, last, EM_REPLACESEL, FALSE, (long)"%duration%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    }
    break;
  }
  return FALSE;
}



EXPORT void FileFilterUserInterface( MINIFILEPARAM *param, HWND parent )
{
  DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_DIALOG), parent, 
		 Dialog, (LPARAM)param);
}


static int CheckID3V2(FILE *fp, char *title, char *artist, char *album,
		      char *year, int *track)
{
  char hdr[10];
  int unsync, exthdr;
  int size, fullsize, ptr;
  char *data;

  year[0] = album[0] = artist[0] = title[0] = 0;
  *track = 0;

  hdr[0]=0;
  if (!fp) return FALSE;
  fseek(fp, 0, SEEK_SET);
  fread(&hdr, 10, 1, fp);
  if (hdr[0]!='I' || hdr[1]!='D' || hdr[2]!='3') return FALSE;
  unsync = hdr[5]&128;
  exthdr = hdr[5]&64;
  if (exthdr||unsync) return FALSE;  // TBD
  fullsize = (hdr[9]&0x7f) | ((hdr[8]&0x7f)<<7) | ((hdr[7]&0x7f)<<14);
  data = (char*)calloc(fullsize, 1);
  fread(data, fullsize, 1, fp);
  ptr = 0;
  while (ptr<fullsize) {
    if (!strncmp(data+ptr, "TRCK", 4)) {
      ptr += 4;
      size = ((*(data+ptr+3))&0x7f) | (((*(data+ptr+2))&0x7f)<<7) |
	(((*(data+ptr+1))&0x7f)<<14);
      ptr += 6;
      strncpy(hdr, 1+data+ptr, min(8, size));
      hdr[min(8, size)-1]=0;
      sscanf(hdr, "%d", track);
      ptr += size;
    } else if (!strncmp(data+ptr, "TYER", 4)) {
      ptr += 4;
      size = ((*(data+ptr+3))&0x7f) | (((*(data+ptr+2))&0x7f)<<7) |
	(((*(data+ptr+1))&0x7f)<<14);
      ptr += 6;
      strncpy(year, 1+data+ptr, min(127, size));
      year[min(127,size)-1]=0;
      ptr += size;
    } else if (!strncmp(data+ptr, "TALB", 4)) {
      ptr += 4;
      size = ((*(data+ptr+3))&0x7f) | (((*(data+ptr+2))&0x7f)<<7) |
	(((*(data+ptr+1))&0x7f)<<14);
      ptr += 6;
      strncpy(album, 1+data+ptr, min(127, size));
      album[min(127,size)-1]=0;
      ptr += size;

    } else if (!strncmp(data+ptr, "TPE1", 4)) {
      ptr += 4;
      size = ((*(data+ptr+3))&0x7f) | (((*(data+ptr+2))&0x7f)<<7) |
	(((*(data+ptr+1))&0x7f)<<14);
      ptr += 6;

      strncpy(artist, 1+data+ptr, min(127, size));
      artist[min(127,size)-1]=0;
      ptr += size;

    } else if (!strncmp(data+ptr, "TIT2", 4)) {
      ptr += 4;
      size = ((*(data+ptr+3))&0x7f) | (((*(data+ptr+2))&0x7f)<<7) | 
	(((*(data+ptr+1))&0x7f)<<14);
      ptr += 6;

      strncpy(title, 1+data+ptr, min(127, size));
      title[min(127,size)-1]=0;
      ptr += size;
    } else {
      ptr += 4;
      size = ((*(data+ptr+3))&0x7f) | (((*(data+ptr+2))&0x7f)<<7) | 
	(((*(data+ptr+1))&0x7f)<<14);
      ptr += 6;
      ptr += size;
    }
  }
  free(data);
  return TRUE;
}	



EXPORT MINIZIPINFO *FileFilterExecute( MINIFILEPARAM *param,
				       MINIZIPINFO *mz,
				       char *label,
				       unsigned int labelSize,
				       char *path )
{
  int j, track;
  unsigned int i;
  FILE *fp;
  char fname[MAX_PATH], buff[1024], title[128], artist[128];
  char album[128], year[128], outBuff[1024];
  MPEGDATA mp3data;
  char *fileMap, *commMap, *dot;

  if (!enabled)
    return mz;

  fileMap = (char *)param->user;
  commMap = fileMap + 256;

  for (i=0; i<mz->numfiles; i++) {
		dot = strrchr(mz->fi[i].name, '.');
		if (!dot) dot = "x";
    if (!stricmp(dot, ".mp3")) {
      sprintf(fname, "%s\\%s", path, mz->fi[i].name);
      buff[0] = 0;
      fp = fopen(fname, "rb");
      if (fp) {
	fseek(fp, -128, SEEK_END);
	fread(buff, 128, 1, fp);
      }
      if (buff[0]=='T' && buff[1]=='A' && buff[2]=='G') {
	if (buff[125]==0) track = buff[126];
	else track = 0;
	strncpy(title, buff+3, 30); title[30]=0;
	strncpy(artist, buff+3+30, 30); artist[30]=0;
	strncpy(album, buff+3+30+30, 30); album[30]=0;
	strncpy(year, buff+3+30+30+30, 4); year[4]=0;
	for (j=29; j>0; j--) if (title[j]==' ') title[j]=0; else j=0;
	for (j=29; j>0; j--) if (artist[j]==' ') artist[j]=0; else j=0;
	for (j=29; j>0; j--) if (album[j]==' ') album[j]=0; else j=0;
	for (j=3; j>0; j--) if (year[j]==' ') year[j]=0; else j=0;
	GetMP3Info(fp, &mp3data);
	if (!title[0] || (title[0]==' '&&title[1]==0))
	  strncpy(title, mz->fi[i].name, 128);
	FormatString(fileMap, mz->fi[i].name, outBuff, 256,
		     title, artist, album, year, track,
		     mp3data.bitRateText, mp3data.durationText);
	strncpy(mz->fi[i].name, outBuff, NAMELEN);
	mz->fi[i].name[NAMELEN-1]=0;
	FormatString(commMap, mz->fi[i].comm, outBuff, 256,
		     title, artist, album, year, track,
		     mp3data.bitRateText, mp3data.durationText);
	strncpy(mz->fi[i].comm, outBuff, COMMLEN);
	mz->fi[i].comm[COMMLEN-1]=0;
      } else {
	if (CheckID3V2(fp, title, artist, album, year, &track)) {
	  if (!title[0] || (title[0]==' '&&title[1]==0))
	    strncpy(title, mz->fi[i].name, 128);
	  GetMP3Info(fp, &mp3data);
	  FormatString(fileMap, mz->fi[i].name, outBuff, 256,
		       title, artist, album, year, track,
		       mp3data.bitRateText, mp3data.durationText);
	  strncpy(mz->fi[i].name, outBuff, NAMELEN);
	  mz->fi[i].name[NAMELEN-1]=0;

	  FormatString(commMap, mz->fi[i].comm, outBuff, 256,
		       title, artist, album, year, track,
		       mp3data.bitRateText, mp3data.durationText);
	  strncpy(mz->fi[i].comm, outBuff, COMMLEN);
	  mz->fi[i].comm[COMMLEN-1]=0;
	} else if (GetMP3Info(fp, &mp3data)) {
	  FormatString(commMap, mz->fi[i].comm, outBuff, 256,
		       title, artist, album, year, track,
		       mp3data.bitRateText, mp3data.durationText);
	  strncpy(mz->fi[i].comm, outBuff, COMMLEN);
	  mz->fi[i].comm[COMMLEN-1]=0;
	}
      }
      if (fp) fclose(fp);

    }
    if (mz->fi[i].zi) {
      sprintf(fname, "%s\\%s", path, mz->fi[i].name);
      FileFilterExecute( param, mz->fi[i].zi, label, labelSize, fname);
    }
  }
  return mz;
}

EXPORT void FileFilterUninitialize( MINIFILEPARAM *param )
{
  RegSetValueEx(param->userKey, "string", 0, REG_SZ,
		(unsigned char*)param->user, 256);
  RegSetValueEx(param->userKey, "commstring", 0, REG_SZ,
		(unsigned char*)param->user+256, 256);
  RegSetValueEx(param->userKey, "position", 0, REG_BINARY,
		(unsigned char*)&position, sizeof(POINT));
  RegSetValueEx(param->userKey, "enabled", 0, REG_DWORD,
		(unsigned char*)&enabled, sizeof(DWORD));
  free(param->user);
}


